import os
import pathlib

from typing import Union

from xappt.utilities.command_runner import CommandRunner

from xappt.config import log as logger


def is_path_repository(path: Union[str, pathlib.Path]) -> bool:
    """ Check if a path contains a git repository. """
    return CommandRunner(cwd=path).run(("git", "rev-parse")).result == 0


def commit_id(path: Union[str, pathlib.Path], *, short: bool = False) -> str:
    """ Fetch the current commit id (optionally the short version) from a git
     repository located at `path`. """
    command = ["git", "rev-parse"]
    if short:
        command.append("--short")
    command.append("HEAD")
    output = CommandRunner(cwd=path).run(command)
    if output.result == 0:
        return output.stdout.strip()
    raise RuntimeError(output.stderr)


def is_dirty(path: Union[str, pathlib.Path]) -> bool:
    cmd = CommandRunner(cwd=path)

    # are we a git repository?
    if not is_path_repository(path):
        logger.warning(f"Not a git repository: {path}")
        return True

    cmd.run(("git", "fetch", "origin"))

    # check for local change
    result = cmd.run(("git", "diff-index", "--quiet", "HEAD", "--"))
    if result.result != 0:
        logger.warning(f"Local changes detected: {path}")
        return True

    # check for remote change
    result = cmd.run(("git", "diff", "--quiet", "master", "origin/master"))
    if result.result != 0:
        logger.warning(f"Remote changes detected: {path}")
        return True

    return False



from __future__ import annotations

import weakref

from collections import defaultdict
from typing import Callable, Optional

__all__ = ["Callback"]


class Callback:
    """ A simple object to run callback functions. Note that adding, removing,
    or clearing the set of functions is deferred until the next time `invoke` is
    called. This ensures that we're not modifying the callback set while we're
    iterating through it.
    """

    def __init__(self):
        self._callback_functions: set[Callable] = set()
        self._deferred_operations: defaultdict[str, set[Optional[Callable]]] = defaultdict(set)
        self._paused = False

    def add(self, cb: Callable):
        self._deferred_operations['add'].add(cb)

    def remove(self, cb: Callable):
        self._deferred_operations['remove'].add(cb)

    def clear(self):
        self._deferred_operations['clear'].add(None)

    def _run_deferred_ops(self):
        """ Process any queued operations for `self._callback_functions`. The
        methods `add`, `remove`, and `clear` build the `self._deferred_operations`
        dictionary which has the following format:

        {
            'add': {callable1, callable2},
            'remove': {callable3, callable4},
            'clear': {None},
        }

        When iterating through this dictionary we look up the key as an
        attribute of the `self._callback_functions` set, and then pass
        each callable as an argument to that function. With the above
        dictionary this is equivalent to:

        self._callback_functions.add(callable1)
        self._callback_functions.add(callable2)
        self._callback_functions.remove(callable3)
        self._callback_functions.remove(callable4)
        self._callback_functions.clear()

        """
        while self._deferred_operations:
            op_name, callables = self._deferred_operations.popitem()
            op_method = getattr(self._callback_functions, op_name)
            for fn in callables:
                if fn is None:
                    op_method()
                else:
                    try:
                        ref = weakref.WeakMethod(fn)
                    except TypeError:
                        op_method(weakref.ref(fn))
                    else:
                        op_method(ref)
        self._clear_dead_refs()

    def _clear_dead_refs(self):
        refs_to_remove = set()
        for fn in self._callback_functions:
            if fn() is None:
                refs_to_remove.add(fn)
        for fn in refs_to_remove:
            self._callback_functions.remove(fn)

    def invoke(self, *args, **kwargs):
        self._run_deferred_ops()
        if self._paused:
            return
        for fn in self._callback_functions:
            fn()(*args, **kwargs)

    @property
    def paused(self) -> bool:
        return self._paused

    @paused.setter
    def paused(self, value: bool):
        self._paused = value



from xappt.utilities.path.misc import find_files, get_unique_name
from xappt.utilities.path.temp_path import temp_path
from xappt.utilities.path.misc import search_files, unique_path, UniqueMode, user_data_path
from xappt.utilities.path.temp_path import temporary_path



from __future__ import annotations

from typing import TYPE_CHECKING

from xappt.models.parameter.base import BaseParameterPlugin
if TYPE_CHECKING:
    from xappt.models.plugins.interface import BaseInterface


class BaseTool(BaseParameterPlugin):
    def __init__(self, *, interface: BaseInterface, **kwargs):
        super(BaseTool, self).__init__(**kwargs)
        self.interface = interface

    @classmethod
    def collection(cls) -> str:
        return "tool"

    def execute(self, **kwargs) -> int:
        raise NotImplementedError

    def abort_requested(self):
        if self.interface.command_runner.running:
            self.interface.command_runner.abort()



import xappt

TOOL_CHOICES = {
    "data types": "datatypes",
    "modify choices": "modifychoices",
    "none (quit)": None,
}


@xappt.register_plugin
class ToolChaining(xappt.BaseTool):
    next_tool = xappt.ParamString(description="Choose the next task", choices=list(TOOL_CHOICES.keys()))
    persistent_data = xappt.ParamString(description="This value should persist between tools")

    @classmethod
    def name(cls) -> str:
        return "chaining-example"

    @classmethod
    def help(cls) -> str:
        return "Invoking a dynamic chain of tools"

    @classmethod
    def collection(cls) -> str:
        return "Examples"

    def execute(self, **kwargs) -> int:
        self.interface.tool_data['persistent_data'] = self.persistent_data.value

        first_run = kwargs.get('first_run', True)
        if first_run:
            self.interface.write_stdout("Any queued tools will be initialized with the "
                                        "contents of `interface.tool_data`.")
            self.interface.write_stdout("`interface.tool_data` will also be passed to `BaseTool.execute` as **kwargs.")
            self.interface.write_stdout("This can be used not only to initialize subsequent")
            self.interface.tool_data['first_run'] = False

        next_tool_name = TOOL_CHOICES[self.next_tool.value]
        if next_tool_name is not None:
            self.interface.add_tool(next_tool_name)
            self.interface.add_tool(ToolChaining)  # return here after the other tool runs

        return 0



import pprint
import time

import xappt


@xappt.register_plugin
class DataTypes(xappt.BaseTool):
    integer_choice = xappt.ParamInt(description="Choose an item", choices=("apple", "banana", "cantaloupe"))
    integer_range = xappt.ParamInt(description="A number between 1 and 20, inclusive", minimum=1, maximum=20, default=1)
    hidden_parameter = xappt.ParamInt(description="Any integer", hidden=True, default=7)
    boolean_check = xappt.ParamBool(description="Toggle me")
    string_field = xappt.ParamString(description="Enter any text")
    string_choice = xappt.ParamString(description="Choose an item", choices=("red", "green", "blue"))
    list_choice = xappt.ParamList(description="Choose multiple items", choices=("first", "second", "third", "fourth"))

    @classmethod
    def help(cls) -> str:
        return "An example of the various parameter data types"

    @classmethod
    def collection(cls) -> str:
        return "Examples"

    def execute(self, **kwargs) -> int:
        self.interface.progress_start()
        for i in range(1, 101):
            self.interface.progress_update(f"Working {i}%", i / 100.0)
            time.sleep(0.01)
        self.interface.progress_end()

        param_dict_string = pprint.pformat(self.param_dict(), indent=2, sort_dicts=False)
        self.interface.message(param_dict_string)

        return 0



from typing import Generator

from xappt.models.parameter.meta import ParamMeta
from xappt.models.parameter.model import Parameter, ParamSetupDict, ParameterDescriptor
from xappt.models.parameter.errors import ParameterValidationError
from xappt.models.plugins.base import BasePlugin


class BaseParameterPlugin(BasePlugin, metaclass=ParamMeta):
    def __init__(self, **kwargs):
        super().__init__()
        for param_name in self._parameters_:
            param: Parameter = getattr(self, param_name)
            if param_name in kwargs:
                param_value = kwargs[param_name]
                if isinstance(param_value, ParamSetupDict):
                    param.update(param_value)
                elif param_value is None:
                    pass  # leave this parameter's value empty
                else:
                    param._value = param.validate(param_value)
            else:
                try:
                    param._value = param.validate(param.value)
                except ParameterValidationError:
                    # run validations, but don't raise validation errors
                    pass

    @classmethod
    def class_parameters(cls) -> Generator[ParameterDescriptor, None, None]:
        for item in cls._parameters_:
            yield getattr(cls, item)

    def parameters(self) -> Generator[Parameter, None, None]:
        for item in self._parameters_:
            yield getattr(self, item)

    def param_dict(self) -> dict:
        d = {}
        for p in self.parameters():
            d[p.name] = p.value
        return d

    def validate(self):
        for param in self.parameters():
            param.validate(param.value)



from __future__ import annotations

from typing import Any, Optional, Sequence, Type
from typing import TYPE_CHECKING

from xappt.models.callback import Callback
if TYPE_CHECKING:
    from xappt.models.parameter.validators import BaseValidator


class ParamSetupDict(dict):
    def __init__(self, default=None, required=None, choices=None, options=None, value=None,
                 metadata=None, hidden=None, **kwargs):
        kwargs['default'] = default
        kwargs['required'] = required
        kwargs['choices'] = choices
        kwargs['options'] = options or {}
        kwargs['value'] = value
        kwargs['metadata'] = metadata or {}
        kwargs['hidden'] = hidden or False
        super(ParamSetupDict, self).__init__(**kwargs)


class Parameter:
    def __init__(self, name: str, *, data_type: Type, **kwargs):
        self.name: str = name
        self.data_type: Type = data_type
        self.description: str = kwargs.get('description', "")
        self.default: Any = kwargs['default'] or data_type()
        self.required: bool = kwargs['required']
        self._choices: Optional[Sequence] = kwargs.get('choices')
        self._options: dict[str: Any] = kwargs.get('options', {})
        self.validators: list[BaseValidator] = []
        self._value = kwargs['value']
        self.metadata: dict = kwargs.get('metadata', {})
        self._hidden: bool = kwargs.get('hidden', False)

        for validator in kwargs.get('validators', []):
            if isinstance(validator, tuple):
                validator, *args = validator
            else:
                args = []
            self.validators.append(validator(self, *args))

        self.on_value_changed = Callback()
        self.on_choices_changed = Callback()
        self.on_options_changed = Callback()
        self.on_visibility_changed = Callback()

    def update(self, update_args: ParamSetupDict):
        self.default = update_args.get('default', self.default)
        self.required = update_args.get('required', self.required)
        self._choices = update_args.get('choices', self._choices)
        self._options.update(update_args.get('options', {}))
        self._value = self.validate(update_args.get('value', self.value))
        self.metadata.update(update_args.get('metadata', {}))

    def validate(self, value: Any) -> Any:
        for validator in self.validators:
            value = validator.validate(value)
        return value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, new_value):
        self._value = new_value
        self.on_value_changed.invoke(param=self)

    @property
    def choices(self) -> Optional[Sequence]:
        return self._choices

    @choices.setter
    def choices(self, new_choices: Optional[Sequence]):
        self._choices = new_choices
        if self.data_type is str:
            if self.default is not None and self.default not in self._choices:
                self.default = self._choices[0]
            if self._value not in self._choices:
                self._value = self.default
        elif self.data_type is int:
            if self.default < 0 or self.default >= len(self._choices):
                self.default = 0
            if self._value < 0 or self._value >= len(self._choices):
                self._value = self.default
        self.on_choices_changed.invoke(param=self)

    @property
    def options(self) -> dict:
        return self._options

    def option(self, key: str, default: Any) -> Any:
        return self._options.get(key, default)

    def set_option(self, key: str, value: Any):
        self._options[key] = value
        self.on_options_changed.invoke(param=self)

    @property
    def hidden(self):  # read only
        return self._hidden

    @hidden.setter
    def hidden(self, new_hidden: bool):
        self._hidden = new_hidden
        self.on_visibility_changed.invoke(param=self)


class ParameterDescriptor:
    __counter = 0

    def __init__(self, data_type: Type, **kwargs):
        cls = self.__class__
        self.storage_name = f"_{cls.__name__}#{cls.__counter}"

        self.param_setup_args = ParamSetupDict(**kwargs)
        self.param_setup_args['data_type'] = data_type

        if "description" not in kwargs:
            self.param_setup_args['description'] = ""

        cls.__counter += 1

    def get_parameter(self, instance):
        param = getattr(instance, self.storage_name, None)
        if param is None:
            # Create new parameter instance of descriptor properties
            param = Parameter(**self.param_setup_args)
            setattr(instance, self.storage_name, param)
            return param
        return param

    def __get__(self, instance, owner):
        if instance is None:
            # Call was not through an instance
            return self
        else:
            return self.get_parameter(instance)

    def __set__(self, instance, value):
        raise AttributeError("Parameter is read only. Use Parameter's value attribute")



from .model import ParameterDescriptor


class ParamMeta(type):
    def __new__(mcs, name, bases, attrs):
        cls = type.__new__(mcs, name, bases, attrs)
        base_parameters = []
        for base in bases:
            try:
                base_parameters += base._parameters_  # noqa
            except AttributeError:
                pass
        cls._parameters_ = base_parameters
        for var_name, var_value in vars(cls).items():
            if isinstance(var_value, ParameterDescriptor):
                try:
                    _ = cls._parameters_.index(var_name)
                except ValueError:
                    cls._parameters_.append(var_name)
                var_value.param_setup_args['name'] = var_name
        return cls
